"""Fragments of data connection."""

DATA_CONNECTION_FRAGMENT = """
    id
    dataDifferencesSummary {added removed total}
    isChecking
    isApplyingDataDifferences
    numberOfAssets
    selectedFolders
    projectId
"""
